#include <paging.h>
#include <screen.h>
#include <interrupts.h>
#include <common.h>

#define TOTAL_PAGES	1024
#define PAGES_LOWER 256
#define KERNEL_PAGES (PAGES_LOWER + k_exe + 1 /* for paging dir */ + 1 /* kernel table */)
extern u32int end;

u32int mem_end; // it is rounded down if it is not page aligned

u32int placement_address = ((u32int) &end);

page_dir_t *dir = 0;
page_table_t *kernel_table = 0;
/* For future reference
	There is only ONE page directory
		it contains 1024 entries
		
	There are 1024 page tables
	each one belongs to a different process
	
*/

extern void enable_paging(u32int dir);
void pageFaultHandler(registers_t regs);

void pageFaultHandler(registers_t regs)
{
	// Do nothing but m_puts and freeze for the moment
	m_puts("Page Fault!\n");
	for(;;);
}

void set_11_paging(page_table_t* table, u32int mem_end) // for kernel table
{
	memset(table, 0, sizeof(page_table_t) * 1024);
	/* This is for pages that are guaranteed to be there.
		list:
			seg1:
				0x00000000
				0x000FFFFF
	*/
	int i = 0;
	u32int placement_aligned = placement_address;
	
	extern u32int start;
	
	u32int k_exe = &end - &start; // find out how many pages need to be allocated on 1:1 paging for kernel
	// page-align it
	k_exe /= 0x1000;
	k_exe ++;
	// executable is k_exe pages long	
	while(i < PAGES_LOWER)
	{
		table->table[i].address = (i * 0x1000);
		table->table[i].global 		= 0; // for now
		table->table[i].alwayszero  = 0;
		table->table[i].dirty		= 0; // NOTE: may cause problems at setting
		table->table[i].accessed	= 1;
		table->table[i].cachedisable= 1;
		table->table[i].writethrough= 0;
		table->table[i].user		= 0;
		table->table[i].rw			= 1;
		table->table[i].present		= 1;
		i++;
	}
	
	while(i < mem_end/0x1000)
	{
		table->table[i].address 	= (i * 0x1000);
		table->table[i].global 		= 0;
		table->table[i].alwayszero	= 0;
		table->table[i].dirty		= 0;
		table->table[i].accessed	= 0;
		table->table[i].cachedisable= 0;
		table->table[i].writethrough= 0;
		table->table[i].user		= 0;
		table->table[i].rw			= 1;
		table->table[i].present		= 1;
		i++;
	}
}

void program_dirent(int num, page_table_t* table, int rw, int user, int cachedisable, int wt)
{
	u32int table_addr = (u32int) table;
	if(table_addr % 0x1000)
	{
		m_puts("ERROR: table is not page-aligned\n");
		for(;;);
	}
	
	dir->table[num].address		= table_addr;
	dir->table[num].ignored		= 0;
	dir->table[num].size		= 0;
	dir->table[num].alwayszero	= 0;
	dir->table[num].accessed		= 0;
	dir->table[num].cachedisable  = (cachedisable) ? 1 : 0;
	dir->table[num].writethrough  = (wt) ? 1 : 0;
	dir->table[num].user			= (user) ? 1 : 0;
	dir->table[num].rw			= (rw) ? 1 : 0;
	dir->table[num].present		= 1;
	
	return;
}

void init_paging(u32int mem_in_kb)
{
	u32int mem_in_b = mem_in_kb * 1024;
	
	m_printx(placement_address);
	placement_address &= 0xFFFFF000;
	placement_address += 0x1000;
	m_printx(placement_address);
	dir = placement_address;
	placement_address += 0x1000;
	kernel_table = (page_table_t *) placement_address;
	mem_in_b &= 0xFFFFF000;
	
	memset(dir, 0, sizeof(page_dir_t));
	
	program_dirent(0,kernel_table,1,0,0,0);
	set_11_paging(kernel_table, mem_in_b);
	
	register_interrupt_handler(14, &pageFaultHandler);
	
	enable_paging((u32int) dir);
}
